// Copyright (C) 2003 by Alexander Motzkau.
// 
// This file is released under the GPL.

#include <stdio.h>
#include <string.h>
#include <sys/ioctl.h>
#include <libdevmapper.h>

// Because we should not include kernel headers like <linux/fs.h>
#define BLKGETSIZE _IO(0x12,96)

char *hex = "0123456789abcdef";

int crypt_swap(char* device, char* name, char *cipher)
{
    char pars[256];
    unsigned char key[32];
    unsigned char rkey[32];
    struct dm_task* dmt;
    unsigned long size;
    int i, len;
    FILE *dev;
    
    len = strlen(cipher);
    if(len>128)
    {
	fprintf(stderr, "Cipher name is too long.\n");
	return 1;
    }
    strcpy(pars, cipher);
    strcat(pars, " "); len++;
    
    if(strlen(device)>255-len-64-5)
    {
	fprintf(stderr, "Device name is too long.\n");
	return 1;
    }
    
    dev = fopen(device,"rb");
    
    if(dev==NULL)
    {
	perror(device);
	// fprintf(stderr, "Couldn't open the device file.\n");
	return 1;
    }
    fread(key,32,1,dev);
    
    if(ioctl(fileno(dev),BLKGETSIZE, &size)==-1)
    {
	perror("Get block size");
	fclose(dev);
	return 1;
    }
    
    fclose(dev);
    
    dev = fopen("/dev/random","rb");
    if(dev!=NULL)
    {
	fread(rkey, 32, 1, dev);
	fclose(dev);
	for(i=0;i<32;i++)
	    key[i]^=rkey[i];
    }
    
    for(i=0;i<32;i++)
    {
	pars[len++]=hex[key[i]>>4];
	pars[len++]=hex[key[i]&15];
    }
    memcpy(pars+len," 0 ",3); len+=3;
    memcpy(pars+len,device,strlen(device)+1);
    strcat(pars+len," 0");
    
    if(!(dmt=dm_task_create(DM_DEVICE_CREATE)))
    {
	fprintf(stderr, "Couldn't create device mapper task.\n");
	return 1;
    }
    
    if(!dm_task_set_name(dmt, name))
    {
	dm_task_destroy(dmt);
	fprintf(stderr, "Couldn't set task name.\n");
	return 1;
    }
    
    if (!dm_task_add_target(dmt, 0, size, "crypt", pars))
    {
	dm_task_destroy(dmt);
	fprintf(stderr, "Couldn't add cipher.\n");
	return 1;
    }
    
    if (!dm_task_run(dmt))
    {
	dm_task_destroy(dmt);
	fprintf(stderr, "Couldn't run task.\n");
	return 1;
    }

    dm_lib_release();
    dm_lib_exit();

    return 0;
}

int usage(FILE* out)
{
    fprintf(out, "Usage: cryptswap <device-path> <new-name> <cipher>\n");
    return 1;
}

int main(int argc, char *argv[])
{
    if(argc!=4)
	return usage(stderr);
    
    return crypt_swap(argv[1], argv[2], argv[3]);
}
